;; SBCL implementation (Cf. the SB-MOP package).

(defclass binary-function (standard-generic-function)
  ()
  (:metaclass sb-mop:funcallable-standard-class))

(defmacro defbinary (function-name lambda-list &rest options)
  `(defgeneric ,function-name ,lambda-list
     (:generic-function-class binary-function)
     ,@options))

(defmethod add-method :before ((bf binary-function) method)
  (let ((method-specializers (sb-mop:method-specializers method)))
    ;; we're adding a binary method
    (assert (= (length method-specializers) 2))
    ;; we're specializing both arguments on the same class
    (assert (apply #'eq method-specializers))))


;; #### NOTE: when specializing COMPUTE-APPLICABLE-METHODS-USING-CLASSES, SBCL
;; #### requires an explicit specialization of COMPUTE-APPLICABLE-METHODS as
;; #### well, even if it does nothing special.

(defmethod sb-mop:compute-applicable-methods-using-classes :before
    ((bf binary-function) classes)
  ;; we're calling a binary function
  (assert (= (length classes) 2))
  ;; we're calling it with both arguments of the same class
  (assert (apply #'eq classes)))

(defmethod sb-mop:compute-applicable-methods :before
    ((bf binary-function) arguments))

(defmethod sb-mop:compute-applicable-methods-using-classes :around
    ((bf binary-function) classes)
  (multiple-value-bind (methods ok) (call-next-method)
    (when ok
      (assert methods)
      (loop :for class :in (sb-mop:class-precedence-list (car classes))
	    :until (eq class (find-class 'standard-object))
	    :do (assert
		 (find-if
		  (lambda (method)
		    (let ((qualifiers (sb-mop:method-qualifiers method)))
		      (and (equal (sb-mop:method-specializers method)
				  (list class class))
			   (not (member :around qualifiers))
			   (not (member :before qualifiers))
			   (not (member :after qualifiers)))))
		  methods))))
    (values methods ok)))

(defmethod sb-mop:compute-applicable-methods ((bf binary-function) arguments)
  (call-next-method))
